#!/usr/bin/python

# Author:		Radomir Mastalerz
# Email:		radekmas@tlen.pl
# Program:	Maximum entropy tagger for polish

import codecs
import sys
import os
import shutil
import re						#regular expressions
import time
import cPickle
import time
import pprint

import nltk.classify.maxent

class maxent:

	attr = dict()
	attr['number'] = ['sg','pl']
	attr['case'] = ['nom', 'gen', 'dat', 'acc', 'inst', 'loc', 'voc']
	attr['gender'] = ['m1', 'm2', 'm3', 'f', 'n']
	attr['person'] = ['pri', 'sec', 'ter']
	attr['degree'] = ['pos', 'com', 'sup']
	attr['aspect'] = ['imperf', 'perf']
	attr['negation'] = ['aff', 'neg']
	attr['accommodability'] = ['congr', 'rec']
	attr['accentability'] = ['akc', 'nakc']
	attr['post-prepositionality'] = ['npraep', 'praep']
	attr['agglutination'] = ['agl', 'nagl']
	attr['vocalicity'] = ['nwok', 'wok']
	attr['fullstoppedness'] = ['pun', 'npun']
	
	attrOrder = ['number','case','gender','person','degree','aspect','negation','accommodability','accentability','post-prepositionality','agglutination','vocalicity','fullstoppedness']

	pos = dict()
	pos['adja'] = []
	pos['adjp'] = []
	pos['adjc'] = []
	pos['conj'] = []
	pos['comp'] = []
	pos['interp'] = []
	pos['pred'] = []
	pos['xxx'] = []
	pos['adv'] = []
	pos['imps'] = ['aspect']
	pos['inf'] = ['aspect']
	pos['pant'] = ['aspect']
	pos['pcon'] = ['aspect']
	pos['qub'] = []
	pos['prep'] = ['case']
	pos['siebie'] = ['case']
	pos['subst'] = ['number', 'case', 'gender']
	pos['depr'] = ['number', 'case', 'gender']
	pos['ger'] = ['number', 'case', 'gender', 'aspect', 'negation']
	pos['ppron12'] = ['number', 'case', 'gender', 'person']
	pos['ppron3'] = ['number', 'case', 'gender', 'person']
	pos['num'] = ['number', 'case', 'gender', 'accommodability']
	pos['numcol'] = ['number', 'case', 'gender', 'accommodability']
	pos['adj'] = ['number', 'case', 'gender', 'degree']
	pos['pact'] = ['number', 'case', 'gender', 'aspect', 'negation']
	pos['ppas'] = ['number', 'case', 'gender', 'aspect', 'negation']
	pos['winien'] = ['number', 'gender', 'aspect']
	pos['praet'] = ['number', 'gender', 'aspect']
	pos['bedzie'] = ['number', 'person', 'aspect']
	pos['fin'] = ['number', 'person', 'aspect']
	pos['impt'] = ['number', 'person', 'aspect', 'vocalicity']
	pos['aglt'] = ['number', 'person', 'aspect']
	pos['brev'] = ['fullstoppedness']
	pos['burk'] = []
	pos['interj'] = []
	pos['ign'] = []
		
	pos_optional = dict()
	pos_optional['adv'] = ['degree']
	pos_optional['qub'] = ['vocalicity']
	pos_optional['prep'] = ['vocalicity']
	pos_optional['ppron12'] = ['accentability']
	pos_optional['ppron3'] = ['accentability','post-prepositionality']
	pos_optional['praet'] = ['agglutination']

	maxIter = 120

	varList = [ 
		"tagCounter", "ctagCounter", "msdCounter",
		"onePossibleCtag", "onePossibleMsd", 
		"unigramWordCtagCounter", "unigramWordCtagAttrMsdCounter",
		"unigramCtagProb", "unigramMsdProb",
		"bestUnigramCtagForWord", "bestUnigramMsdForWordCtagAttr",
		"ctagTrainingClassesDict", "msdTrainingCtagAttrClassesDict",
		"ctagModels", "msdModels",
		"globalResults",
		'listOfCtagModels', 'listOfMsdModels',
		'listOfRejectedCtagModels', 'listOfRejectedMsdModels']
	########## INIT ##########
	
	def __init__(self,learningDirList = ['.'],testingDir = '.'):
		self.learningDirList = learningDirList
		self.testingDir = testingDir
		learningFileList = list()
		testingFileList = list()
		# create list of "ann_morphosyntax.xml" files
		for learningDir in learningDirList:
			for root, subFolders, files in os.walk(learningDir):
				rootFiles = os.listdir(root)
				if "ann_morphosyntax.xml" in rootFiles: 
					learningFileList.append(os.path.join(root,"ann_morphosyntax.xml"))

		for root, subFolders, files in os.walk(testingDir):
			rootFiles = os.listdir(root)
			if "ann_morphosyntax.xml" in rootFiles: 
				testingFileList.append(os.path.join(root,"ann_morphosyntax.xml"))

		print "NUMBER OF LEARNING FILES: ", len(learningFileList)
		print "NUMBER OF TESTING FILES: ", len(testingFileList)

		# create attrValueToName dictionary
		attrValueToName = dict()
		for (attrib, listOfValues) in self.attr.iteritems():
			for value in listOfValues:
				attrValueToName[value] = attrib
		
		self.learningFileList = learningFileList
		self.testingFileList = testingFileList
		self.attrValueToName = attrValueToName
 		self.globalResults = list()
 		
	########## FEATURES ##########
	# let context consits of t(n-2),t(n-1),w(n-2),w(n-1),w(n),w(n+1),w(n+2)

	def features(self,prevPrev,prev,act,next,nextNext):
		f = dict()
		word = act['word']
		f['word'] = word.lower()
		f['prevTag'] = prev.get('tag',u'')
		f['prevPrevTag'] = prevPrev.get('tag',u'')
		f['prevWord'] = prev.get('word',u'').lower()
		f['isUpper'] = word[0].isupper()			
		f['lastLetter'] = word[(-1)]		
		if len(word) >= 2:					
			f['lastTwoLetters'] = word[(-2):] 
		else:
			f['lastTwoLetters'] = ''
		return f
		
		
	def ctagFeatures(self,prevPrev,prev,act,next,nexNext):
		f = dict()
		word = act['word'].lower()
		f['word'] = word
		f['prevWord'] = prev.get('word','').lower()
		f['prevCtag'] = prev.get('tag',(u'',u''))[0]
		f['prevPrevCag'] = prevPrev.get('tag',(u'',u''))[0]
		f['nextCtagAC'] = next.get('ctagAC',())
		f['lastLetter'] = word[(-1)]
		if len(word) >= 2:					
			f['lastTwoLetters'] = word[(-2):] 
		else:
			f['lastTwoLetters'] = ''
		if len(word) >= 3:					
			f['lastThreeLetters'] = word[(-3):] 
		else:
			f['lastThreeLetters'] = ''	
		return f

	def msdFeatures(self,attrName,prevPrev,prev,act,next,nextNext):
		f = dict()
		word = act['word'].lower()
		f['word'] = word	
		f['prevWord'] = prev.get('word','').lower()
		f['prevCtag'] = prev.get('tag',(u'',u''))[0]
		f['prevPrevCag'] = prevPrev.get('tag',(u'',u''))[0]
		f['nextCtagAC'] = next.get('ctagAC',())
		f['lastLetter'] = word[(-1)]
		if len(word) >= 2:					
			f['lastTwoLetters'] = word[(-2):] 
		else:
			f['lastTwoLetters'] = ''
		if len(word) >= 3:
			f['lastThreeLetters'] = word[(-3):]
		else:
			f['lastThreeLetters'] = ''
		if attrName == 'number':
			f['prevNumber'] = prev.get('msdDcit',dict()).get('number','')
			f['prevPrevNumber'] = prevPrev.get('msdDict',dict()).get('number','')
			f['nextNumberAC'] = next.get('msdAC',dict()).get('number')
		elif attrName == 'case':	
			f['prevCase'] = prev.get('msdDict',dict()).get('case','')			
			f['prevPrevCase'] = prevPrev.get('msdDict',dict()).get('case','')
			f['nextCaseAC'] = next.get('msdAC',dict()).get('case','')
			f['prevPrevIsNot'] = (prevPrev.get('word','').lower() == 'nie')
			f['prevNegation'] = prev.get('msdDict',dict()).get('negation','')
		elif attrName == 'gender':
			f['prevGender'] = prev.get('msdDict',dict()).get('gender','')
			f['prevPrevGender'] = prevPrev.get('msdDict',dict()).get('gender','')
			f['nextGenderAC'] = next.get('msdAC',dict()).get('gender','')			
		elif attrName == 'person':
			f['prevPerson'] = prev.get('msdDict',dict()).get('person','')
			f['prevPrevPerson'] = prevPrev.get('msdDict',dict()).get('person','')
			f['nextPersonAC'] = next.get('msdAC',dict()).get('person','')
		elif attrName == 'degree':
			f['prevDegree'] = prev.get('msdDict',dict()).get('degree','')
			f['prevPrevDegree'] = prevPrev.get('msdDict',dict()).get('degree','')
			f['nextDegreeAC'] = next.get('msdAC',dict()).get('degree','')
		elif attrName == 'aspect':
			f['firstLetter'] = word[0]
			if len(word) >= 2:
				f['firstTwoLetters'] = word[0:2]
			else:
				f['firstTwoLetters'] = ''
			if len(word) >= 3:
				f['firstThreeLetters'] = word[0:3]
			else:
				f['firstThreeLetters'] = ''
		elif attrName == 'negation':
			f['startsWithNot'] = word.startswith('nie')
		elif attrName == 'accommodability':
			f['case'] = act.get('msdDict',dict()).get('case','')
			f['prevCase'] = prev.get('msdDict',dict()).get('case','')
			f['nextCase'] = next.get('msdDict',dict()).get('case','')
		#elif attrName == 'accentability':
		#elif attrName == 'post-prepositionality':
		#elif attrName == 'agglutination':
		#elif attrName == 'vocalicity':	
		elif attrName == 'fullstoppedness':
			f['nextIsDot'] = (next.get('word','') == '.')
			f['nextNextIsUppercase'] = nextNext.get('word',' ')[0].isupper()		
		return f

# 'number','case','gender','person','degree','aspect','negation','accommodability','accentability','post-prepositionality','agglutination','vocalicity','fullstoppedness'

	########## PROCESS TRAINING SET ################
	
	def processTrainingSet(self):
		attr = self.attr
		attrOrder = self.attrOrder
		attrValueToName = self.attrValueToName
		pos = self.pos
		pos_optional = self.pos_optional
		ctagFeatures = self.ctagFeatures
		msdFeatures = self.msdFeatures
		learningFileList = self.learningFileList
		
		segmentCounter = 0	
		onePossibleCtag = 0
		onePossibleMsd = dict()																		# attrName -> count

		ctagCounter = dict()																			# ctag -> count
		msdCounter= dict()																				# attrName -> count
	
		ctagTrainingClassesDict = dict()													# possibleCtagsTuple -> trainingList
		msdTrainingCtagAttrClassesDict = dict()

		unigramWordCtagCounter = dict()														# word -> ctag -> count				
		unigramWordCtagAttrMsdCounter = dict()										# word -> ctag -> attrName -> attrVal -> counter
		
		prevPrev = dict()
		prev = dict()
		act = dict()
		next = dict()
		nextNext = dict()
			
		open = codecs.open
		search = re.search
		getTime = time.time
		
		
		print "########## PROCESSING TRAINING SET ##########"
		t_start = getTime()

		
		for filename in learningFileList:
			parsedSegments = list()
			#print "LEARNING ", filename
			sys.stdout.write('.')
			file = open(filename, "r", encoding='utf-8')
			lines = file.readlines()
			lines.reverse()
			line = lines.pop()
			file.close()
			while len(lines) > 0:
				if line.find("<seg") >= 0:
					segment = dict()
					segmentCounter += 1
					#inside segment
					m = search('<seg xlink:href="(.+?)" xml:id="(.+?)">',line)
					segment['segmId'] = m.group(1)
					segment['morphId'] = m.group(2)
					# <f name = "orth">						
					while line.find('<f name="orth">') < 0:
						line = lines.pop()
					while line.find('<string>') < 0:
						line = lines.pop()
					m = search('<string>(.*?)</string>',line)
					segment['word'] = m.group(1)
					# <f name = "interps">
					while line.find('<f name="interps">') < 0:
						line = lines.pop()			
					# parse lex tags
					lexList = list()
					while True:
						while line.find('<fs type="lex" xml:id=') < 0 and line.find('<f name="disamb">') < 0:
							line = lines.pop()
						if line.find('<f name="disamb">') >= 0:
							break
						else:
							#    <fs type="lex" xml:id="morph_4.52.1-lex">
							#       <f name="base">
							while line.find('<f name="base">') < 0:
								line = lines.pop()
							while line.find('<string>') < 0:
								line = lines.pop()
							m = search('<string>(.*?)</string>',line)
							base = m.group(1)			
							#       <f name="ctag">
							while line.find('<f name="ctag">') < 0:
								line = lines.pop()
							while line.find('<symbol') < 0:
								line = lines.pop()
							m = search('value="(.*?)"',line)
							ctag = m.group(1)	
							#       <f name="msd">	
							while line.find('<f name="msd">') < 0:
								line = lines.pop()
							#			<symbol value=>
							while line.find('</f>') < 0:
								if line.find('<symbol') >= 0:
									m1 = search('value="(.*?)"',line)
									m2 = search('xml:id="(.*?)"',line)
									lexList.append((base,ctag,m1.group(1),m2.group(1)))
								line = lines.pop()
							if len(lexList) < 1:
								print "ERROR: len(lexList) = 0 -> some <symbol> tags might have been not parsed"
					# <f name="disamb">
					while line.find('<f name="disamb">') < 0 and line.find('</seg>') < 0:
						line = lines.pop()
					if line.find('</seg>') >= 0:
						print 'ERROR: <f name="disamb"> not found'				
					while line.find("<f fVal=") < 0:
						line = lines.pop()
					m = search('<f fVal="#(.*?)" name="choice"/>', line)
					tagId = m.group(1)
					while line.find("</seg>") < 0:
						line = lines.pop()
					# we have reached end of segment, find correct tag
					correctTag = ""
					correctCtag = ""
					for base,ctag,msd,msdId in lexList:
						if msdId == tagId:
							correctTag = (ctag,msd)
							correctCtag = ctag
					# find correct msd values for specified attributes
					possibleAttrs = set(pos[correctCtag] + pos_optional.get(correctCtag,[]))
					correctMsd = dict()
					for el in possibleAttrs:					
						correctMsd[el] = u''
					splitMsd = correctTag[1].split(':')			
					for msdVal in splitMsd:
						if msdVal != u'':
							correctMsd[attrValueToName[msdVal]] = msdVal
					# find ambiguity classes
					possibleCtagValues = set()
					possibleMsdValues = dict()
					for (base,ctag,msd,xml_id) in lexList:
						possibleCtagValues.add(ctag)
						if ctag == correctCtag:
							foundAttrs = set()
							splitMsd = msd.split(':')
							for msdVal in splitMsd:
								if msdVal != u'':
									attrName = attrValueToName[msdVal]
									foundAttrs.add(attrName)
									attrSet = possibleMsdValues.setdefault(attrName,set())
									attrSet.add(msdVal)
							for attrName in possibleAttrs.difference(foundAttrs):
								attrSet = possibleMsdValues.setdefault(attrName,set())
								attrSet.add(u'')
					possibleCtagsList = list(possibleCtagValues)
					possibleCtagsList.sort()
					possibleCtagsTuple = tuple(possibleCtagsList)
					if len(possibleCtagValues) == 1:
						onePossibleCtag += 1
					possibleMsdValuesTuples = dict()
					for attrName, attrValuesSet in possibleMsdValues.iteritems():
						attrValuesList = list(attrValuesSet)
						attrValuesList.sort()
						attrValuesTuple = tuple(attrValuesList)
						possibleMsdValuesTuples[attrName] = attrValuesTuple						
						oldCount = msdCounter.setdefault(attrName,0)
						msdCounter[attrName] = oldCount + 1
						if len(attrValuesList) == 1:
							oldCount = onePossibleMsd.setdefault(attrName,0)
							onePossibleMsd[attrName] = oldCount + 1
					#		print "     one possible %s = %s" % (attrName,attrValuesList[0])
					oldCount = ctagCounter.setdefault(correctCtag,0)
					ctagCounter[correctCtag] = oldCount + 1
					# assign values to segment dictionary
					segment['msdDict'] = correctMsd
					segment['lexList'] = lexList
					segment['tag'] = correctTag
					segment['ctagAC'] = possibleCtagsTuple
					segment['msdAC'] = possibleMsdValuesTuples
					# update ctag unigram counter
					wordLowercase = segment['word'].lower()				
					unigramCtagCounter = unigramWordCtagCounter.setdefault(wordLowercase,dict())
					count = unigramCtagCounter.setdefault(correctCtag,0) 
					unigramCtagCounter[correctCtag] = count + 1
					# update unigram counts for msd attributes
					unigramCtagAttrMsdCounter = unigramWordCtagAttrMsdCounter.setdefault(wordLowercase,dict())
					unigramAttrMsdCounter = unigramCtagAttrMsdCounter.setdefault(correctCtag,dict())
					for attr,val in correctMsd.iteritems():
						unigramMsdCounter = unigramAttrMsdCounter.setdefault(attr,dict())
						count = unigramMsdCounter.setdefault(val,0)
						unigramMsdCounter[val] = count + 1
					# compute context if we have reached actual word
					parsedSegments.append(segment)
				else:
					line = lines.pop()
				#if line.find("<seg") >= 0:
			#while

			prevPrev = dict()	
			prev = dict()
			act = dict()
			next = dict()
			nextNext = dict()
			
			parsedSegments.extend([dict(),dict()])
			for segment in parsedSegments:
				prevPrev = prev
				prev = act
				act = next
				next = nextNext
				nextNext = segment
				if len(act) > 0:
					#add context to ctag and msd ambiguity classes
					correctCtag = act['tag'][0]
					correctMsd = act['msdDict']
					ctagAC = act['ctagAC']
					msdAC = act['msdAC']
					# find possible ctag and msd values
					if len(ctagAC) > 1:
						ctagContext = ctagFeatures(prevPrev,prev,act,next,nextNext)
						ctagClass = ctagTrainingClassesDict.setdefault(ctagAC,list())
						ctagClass.append((ctagContext,correctCtag))						
					for attrName, attrValuesTuple in msdAC.iteritems():
						if len(attrValuesTuple) > 1:
							msdContext = msdFeatures(attrName,prevPrev,prev,act,next,nextNext)
							msdTrainingCtagAttrClassesDictGivenCorrect = msdTrainingCtagAttrClassesDict.setdefault(correctCtag,dict())
							classesDict = msdTrainingCtagAttrClassesDictGivenCorrect.setdefault(attrName,dict())
							trainingList = classesDict.setdefault(attrValuesTuple,list())
							trainingList.append((msdContext,correctMsd[attrName]))
					#for
				#if
			#for
			del parsedSegments						#delete list of segments - it's huge		
		#for
		
		t_end = getTime()
		print ""
		print "PROCESSING %d TRAINING FILES TOOK %.2f SECONDS" % (len(learningFileList), t_end-t_start)

		self.ctagCounter = ctagCounter
		self.msdCounter = msdCounter

		self.unigramWordCtagCounter = unigramWordCtagCounter
		self.unigramWordCtagAttrMsdCounter = unigramWordCtagAttrMsdCounter

		self.onePossibleCtag = onePossibleCtag
		self.onePossibleMsd = onePossibleMsd

		#self.ctagTrainingListForGeneralMaxent = ctagTrainingListForGeneralMaxent										# NOT USED AT THE MOMENT
		self.msdTrainingCtagAttrClassesDict = msdTrainingCtagAttrClassesDict
		self.ctagTrainingClassesDict = ctagTrainingClassesDict


	################### CREATE LIST OF ALL MODELS TO TRAIN ###################################

	def createListOfModels(self):

		time_start = time.time()

		attr = self.attr		
		ctagCounter = self.ctagCounter

		unigramWordCtagCounter = self.unigramWordCtagCounter
		unigramWordCtagAttrMsdCounter = self.unigramWordCtagAttrMsdCounter

		msdTrainingCtagAttrClassesDict = self.msdTrainingCtagAttrClassesDict
		ctagTrainingClassesDict = self.ctagTrainingClassesDict

		listOfCtagModels = list()
		listOfMsdModels = list()
		listOfRejectedCtagModels = list()
		listOfRejectedMsdModels = list()
	
		ctagCounterSum = sum(ctagCounter.itervalues())
#		treshold = max(ctagCounterSum / 30000, 10)
		treshold = 20
	
		for (ctagClass,trainingList) in ctagTrainingClassesDict.iteritems():
			if len(trainingList) >= (treshold):
				listOfCtagModels.append((ctagClass,trainingList))
			else:
				#print "REJECTED CTAG " + str(len(trainingList)) +  " / " + str(treshold) + "   " + str(ctagClass)  
				listOfRejectedCtagModels.append((ctagClass,trainingList))
				
		for (ctag, attrClassTrainingDict) in msdTrainingCtagAttrClassesDict.iteritems():
			for (attrib,classTrainingDict) in attrClassTrainingDict.iteritems():
				attribTrainingLen = 0
				for (AC, trainingList) in classTrainingDict.iteritems():
					attribTrainingLen += len(trainingList)
				#attribTrainingTreshold = max(attribTrainingLen / 10000, 10)
				attribTrainingTreshold = 10			
				for (AC, trainingList) in classTrainingDict.iteritems():
					if len(trainingList) >= attribTrainingTreshold:
						listOfMsdModels.append((ctag,attrib,AC,trainingList))
					else:
						listOfRejectedMsdModels.append((ctag,attrib,AC,trainingList))
						#print "REJECTED MSD " + str(len(trainingList)) +  " / " + str(attribTrainingTreshold) + " : " + str(ctag) + " | " + str(attrib) + " | " + str(AC) 

		
		mDir = os.path.join(self.testingDir,"_ctagMaxentModels")
		shutil.rmtree(mDir, ignore_errors=True)
		os.mkdir(mDir)
		mDir = os.path.join(self.testingDir,"_msdMaxentModels")
		shutil.rmtree(mDir, ignore_errors=True)
		os.mkdir(mDir)
		mDir = os.path.join(self.testingDir,"_ctag_most_informative_features")
		shutil.rmtree(mDir, ignore_errors=True)
		os.mkdir(mDir)
		mDir = os.path.join(self.testingDir,"_msd_most_informative_features")
		shutil.rmtree(mDir, ignore_errors=True)
		os.mkdir(mDir)


		time_end = time.time()
		print "CREATING LIST OF %d CTAG AND %d MSD MODELS TOOK %.2f SECONDS" % (len(listOfCtagModels),len(listOfMsdModels),time_end - time_start)
		print "REJECTED CTAG MODELS = %d (TRESHOLD = %d), REJECTED MSD MODELS = %d" % (len(listOfRejectedCtagModels),treshold,len(listOfRejectedMsdModels))

		self.listOfCtagModels = listOfCtagModels
		self.listOfMsdModels = listOfMsdModels
		self.listOfRejectedCtagModels = listOfRejectedCtagModels
		self.listOfRejectedMsdModels = listOfRejectedMsdModels
		
	################# LEARN EVERY N-th MODEL FROM LIST OF CTAG AND MSD MODELS AND SAVE TO FILE #################	
	
	def learnSelectedModels(self,k,n):
		
		listOfCtagModels = self.listOfCtagModels
		listOfMsdModels = self.listOfMsdModels
		maxIter = self.maxIter
		attr = self.attr
		
		getTime = time.time
		write = sys.stdout.write		
		globalStart = getTime()

		ctagDir = os.path.join(self.testingDir,"_ctag_most_informative_features")
		if not os.path.isdir(ctagDir):
			os.mkdir(ctagDir)

		msdRoot = os.path.join(self.testingDir,"_msd_most_informative_features")
		if not os.path.isdir(msdRoot):
			os.mkdir(msdRoot)

		msdDir = dict()
		for attrName in attr:
			actDir = os.path.join(msdRoot, attrName)
			msdDir[attrName] = actDir
			if not os.path.isdir(actDir):
				os.mkdir(actDir)
		
		# build maximum entropy models for various ctag ambiguity classes
		ctagModels = dict()
		ctagModelCounter = 0
		learntCtagModelCounter = 0
		ctagModelsStart = getTime()
		for (ctagClass,trainingList) in listOfCtagModels:
			if (ctagModelCounter%n)== k:
				write('.')
				model = nltk.classify.maxent.MaxentClassifier.train(trainingList,algorithm='IIS',trace=0,max_iter=maxIter)
				ctagModels[ctagClass] = model 
				learntCtagModelCounter += 1
				# write most informative features to file
				filename = ""
				for c in ctagClass:
					if filename == "":
						filename = str(c) 
					else:
						filename = filename + "_" +str(c)	
				f = open(os.path.join(ctagDir,filename),"w")
				sys.stdout = f
				model.show_most_informative_features(10000)
				sys.stdout = sys.__stdout__ 
				f.close()
			ctagModelCounter += 1
		ctagModelsEnd = getTime()
		print ""
		print "LEARNING %d / %d CTAG MAXENT MODELS TOOK %.2f SECONDS" % (learntCtagModelCounter, ctagModelCounter, ctagModelsEnd - ctagModelsStart)

		# build maxium entropy models for different ambiguity classes for gender, case etc.
		msdModels = dict()
		msdModelCounter = 0
		learntMsdModelCounter = 0
		msdModelsStart = getTime()
		for (ctag,attrib,AC,trainingList) in listOfMsdModels:
			if (msdModelCounter%n == k):
				write('.')
				msdModels2 = msdModels.setdefault(ctag,dict())
				msdModels3 = msdModels2.setdefault(attrib,dict())
				model = nltk.classify.maxent.MaxentClassifier.train(trainingList,algorithm='IIS',trace=0,max_iter=maxIter)
				msdModels3[AC] = model
				learntMsdModelCounter += 1
				# write most informative features to file
				filename = str(ctag)
				for c in AC:
					filename = filename + "_" +str(c)
				f = open(os.path.join(msdDir[attrib],filename),"w")
				sys.stdout = f
				model.show_most_informative_features(20000)
				sys.stdout = sys.__stdout__ 
				f.close()				
			msdModelCounter += 1		
		msdModelsEnd = getTime()
		print ""
		print "LEARNING %d / %d MSD MAXENT MODELS TOOK %.2f SECONDS" % (learntMsdModelCounter, msdModelCounter, msdModelsEnd - msdModelsStart) 

		t_start = getTime()
		# SAVE ALL DATA
		mDir = os.path.join(self.testingDir,"_ctagMaxentModels")
		if not os.path.isdir(mDir):
			os.mkdir(mDir)
			
		f = open(os.path.join(mDir,str(k)),"w")
		cPickle.dump(ctagModels,f)
		f.close()

		mDir = os.path.join(self.testingDir,"_msdMaxentModels")
		if not os.path.isdir(mDir):
			os.mkdir(mDir)
		f = open(os.path.join(mDir,str(k)),"w")
		cPickle.dump(msdModels,f)
		f.close()
		
		t_end = getTime()
		print "SAVING ABOVE MODELS TOOK %.2f SECONDS" % (t_end - t_start)
		
		return (ctagModels,msdModels)

	##################### COMBINE RESULTS FROM DISTRIBUTED COMPUTATIONS ##########################
	def combineDistributedModels(self,n):
		
		attr = self.attr
		ctagCounter = self.ctagCounter

		unigramWordCtagCounter = self.unigramWordCtagCounter
		unigramWordCtagAttrMsdCounter = self.unigramWordCtagAttrMsdCounter
		testingDir = self.testingDir
		
		getTime = time.time
		write = sys.stdout.write
		globalStart = getTime()

		ctagModels = dict()
		mDir = os.path.join(testingDir,"_ctagMaxentModels")
		for k in xrange(0,n):	
			if os.path.isfile(os.path.join(mDir,str(k))):			
				f = open(os.path.join(mDir,str(k)),"r")
				loadedDict = cPickle.load(f)
				f.close()	
				for ctagAC,model in loadedDict.iteritems():		
					ctagModels[ctagAC] = model
			else:
				print "WARNING: NO CTAG MODELS FOR FILE MACHINE %d" % k
				
		print "%d CTAG MODELS COMBINED" % len (ctagModels)
		
		
		msdModels = dict()
		mDir = os.path.join(testingDir,"_msdMaxentModels")
		for k in xrange(0,n):	
			if os.path.isfile(os.path.join(mDir,str(k))):			
				f = open(os.path.join(mDir,str(k)),"r")
				loadedDict = cPickle.load(f)
				f.close()
				for ctag,attrTupleDict in loadedDict.iteritems():
					msdModels2 = msdModels.setdefault(ctag,dict())
					for attrName, tupleDict in attrTupleDict.iteritems():
						msdModels3 = msdModels2.setdefault(attrName,dict())
						for tup,model in tupleDict.iteritems():
							msdModels3[tup] = model
			else:
				"WARNING: NO MSD MODELS FOR MACHINE %d" % k
		
		counter = 0
		for ctag,attrTupleDict in msdModels.iteritems():
			for attrName, tupleDict in attrTupleDict.iteritems():
				for tup in tupleDict:
					counter += 1
		
		print "%d MSD MODELS COMBINED" % counter	

		# build unigram probabilities for ctag
		unigramModelsStart = getTime()
		segcount = sum(ctagCounter.itervalues())
		unigramWordCtagProb = dict()
		bestUnigramCtagForWord = dict()
		for word,ctagDict in unigramWordCtagCounter.iteritems():
			wordCount = float(sum(ctagDict.itervalues()))
			unigramCtagProb = unigramWordCtagProb.setdefault(word,dict())
			bestProb = 0.0
			bestCtag = u''
			for ctag,count in ctagDict.iteritems():
				prob = float(count) / wordCount
				unigramCtagProb[ctag] = prob
				if prob > bestProb:
					bestProb = prob
					bestCtag = ctag
			bestUnigramCtagForWord[word] = bestCtag
		
		# build unigram probabilities for 
		unigramWordCtagAttrMsdProb = dict()
		bestUnigramMsdForWordCtagAttr = dict()
		for word,ctagAttrMsdCounter in unigramWordCtagAttrMsdCounter.iteritems():
			unigramCtagAttrMsdProb = unigramWordCtagAttrMsdProb.setdefault(word,dict())
			bestUnigramMsdForCtagAttr = bestUnigramMsdForWordCtagAttr.setdefault(word,dict())
			wordCount = float(sum(unigramWordCtagCounter[word].itervalues()))
			for ctag,attrMsdCounter in ctagAttrMsdCounter.iteritems():
				unigramAttrMsdProb = unigramCtagAttrMsdProb.setdefault(ctag,dict())
				bestUnigramMsdForAttr = bestUnigramMsdForCtagAttr.setdefault(ctag,dict())
				ctagCount = unigramWordCtagAttrMsdCounter[word][ctag]
				for attr,msdCounter in attrMsdCounter.iteritems():
					unigramMsdProb = unigramAttrMsdProb.setdefault(attr,dict())
					attrCount = sum(msdCounter.itervalues())
					bestProb = 0.0
					bestMsdVal = u''
					for msdVal,count in msdCounter.iteritems():
						prob = float(count) / float(attrCount) 
						unigramMsdProb[msdVal] = prob
						if prob > bestProb:
							bestProb = prob
							bestMsdVal = msdVal
					bestUnigramMsdForAttr[attr] = bestMsdVal
		unigramModelsEnd = getTime()
		print "LEARNING 2 UNIGRAM MODLES TOOK %.2f SECONDS" % (unigramModelsEnd - unigramModelsStart)
	
				
		# ASSIGN VARIABLES TO OBJECT
		
		self.unigramWordCtagProb = unigramWordCtagProb
		self.unigramWordCtagAttrMsdProb = unigramWordCtagAttrMsdProb

		self.ctagModels = ctagModels
		self.msdModels = msdModels

		self.bestUnigramCtagForWord = bestUnigramCtagForWord
		self.bestUnigramMsdForWordCtagAttr = bestUnigramMsdForWordCtagAttr

	
	##################### LEARN MAXIMUM ENTROPY AND UNIGRAM CLASSIFIERS ###########################

	def learn(self):
		print "########## LEARNING CLASSIFIERS ##########"

		attr = self.attr
		maxIter = self.maxIter
		ctagCounter = self.ctagCounter

		unigramWordCtagCounter = self.unigramWordCtagCounter
		unigramWordCtagAttrMsdCounter = self.unigramWordCtagAttrMsdCounter

		msdTrainingCtagAttrClassesDict = self.msdTrainingCtagAttrClassesDict
		ctagTrainingClassesDict = self.ctagTrainingClassesDict
		
		getTime = time.time
		write = sys.stdout.write		
		globalStart = getTime()

		# build maximum entropy models for various ctag ambiguity classes
		ctagModels = dict()
		ctagModelCounter = 0
		ctagModelsStart = getTime()
		for (ctagClass,trainingList) in ctagTrainingClassesDict.iteritems():
			write('.')
			ctagModels[ctagClass] = nltk.classify.maxent.MaxentClassifier.train(trainingList,algorithm='IIS',trace=2,max_iter=maxIter)
			ctagModelCounter += 1
		ctagModelsEnd = getTime()
		print ""
		print "LEARNING %d CTAG MAXENT MODELS TOOK %.2f SECONDS" % (ctagModelCounter, ctagModelsEnd - ctagModelsStart)

			
		# build maxium entropy models for different ambiguity classes for gender, case etc.
		msdModels = dict()
		msdModelCounter = 0
		msdModelsStart = getTime()
		for (ctag, attrClassTrainingDict) in msdTrainingCtagAttrClassesDict.iteritems():
			msdModels2 = msdModels.setdefault(ctag,dict())
			for (attrib,classTrainingDict) in attrClassTrainingDict.iteritems():
				msdModels3 = msdModels2.setdefault(attrib,dict())
				for (AC, trainingList) in classTrainingDict.iteritems():
					write('.')
					msdModels3[AC] = nltk.classify.maxent.MaxentClassifier.train(trainingList,algorithm='IIS',trace=2,max_iter=maxIter)
					msdModelCounter += 1
		msdModelsEnd = getTime()
		print ""
		print "LEARNING %d MSD MAXENT MODELS TOOK %.2f SECONDS" % (msdModelCounter, msdModelsEnd - msdModelsStart)
		
		# build unigram probabilities for ctag
		unigramModelsStart = getTime()
		write('.')
		segcount = sum(ctagCounter.itervalues())
		unigramWordCtagProb = dict()
		bestUnigramCtagForWord = dict()
		for word,ctagDict in unigramWordCtagCounter.iteritems():
			wordCount = float(sum(ctagDict.itervalues()))
			unigramCtagProb = unigramWordCtagProb.setdefault(word,dict())
			bestProb = 0.0
			bestCtag = u''
			for ctag,count in ctagDict.iteritems():
				prob = float(count) / wordCount
				unigramCtagProb[ctag] = prob
				if prob > bestProb:
					bestProb = prob
					bestCtag = ctag
			bestUnigramCtagForWord[word] = bestCtag
		
		# build unigram probabilities for 
		unigramWordCtagAttrMsdProb = dict()
		bestUnigramMsdForWordCtagAttr = dict()
		for word,ctagAttrMsdCounter in unigramWordCtagAttrMsdCounter.iteritems():
			unigramCtagAttrMsdProb = unigramWordCtagAttrMsdProb.setdefault(word,dict())
			bestUnigramMsdForCtagAttr = bestUnigramMsdForWordCtagAttr.setdefault(word,dict())
			wordCount = float(sum(unigramWordCtagCounter[word].itervalues()))
			for ctag,attrMsdCounter in ctagAttrMsdCounter.iteritems():
				unigramAttrMsdProb = unigramCtagAttrMsdProb.setdefault(ctag,dict())
				bestUnigramMsdForAttr = bestUnigramMsdForCtagAttr.setdefault(ctag,dict())
				ctagCount = unigramWordCtagAttrMsdCounter[word][ctag]
				for attr,msdCounter in attrMsdCounter.iteritems():
					unigramMsdProb = unigramAttrMsdProb.setdefault(attr,dict())
					attrCount = sum(msdCounter.itervalues())
					bestProb = 0.0
					bestMsdVal = u''
					for msdVal,count in msdCounter.iteritems():
						prob = float(count) / float(attrCount) 
						unigramMsdProb[msdVal] = prob
						if prob > bestProb:
							bestProb = prob
							bestMsdVal = msdVal
					bestUnigramMsdForAttr[attr] = bestMsdVal
		unigramModelsEnd = getTime()
		print "LEARNING 2 UNIGRAM MODLES TOOK %.2f SECONDS" % (unigramModelsEnd - unigramModelsStart)
	
				
		# ASSIGN VARIABLES TO OBJECT
		
		self.unigramWordCtagProb = unigramWordCtagProb
		self.unigramWordCtagAttrMsdProb = unigramWordCtagAttrMsdProb

		self.ctagModels = ctagModels
		self.msdModels = msdModels

		self.bestUnigramCtagForWord = bestUnigramCtagForWord
		self.bestUnigramMsdForWordCtagAttr = bestUnigramMsdForWordCtagAttr
		
	################### TESTING #####################################

	def test(self,filename):
		sys.stdout.write('.')
		file = codecs.open(filename, "r", encoding='utf-8')
		lines = file.readlines()
		lines.reverse()
		if len(lines) < 1:
			print "ERROR: file ", filename, " is empty"
		line = lines.pop()


		pos = self.pos
		pos_optional = self.pos_optional
		attr = self.attr
		attrOrder = self.attrOrder
		attrValueToName = self.attrValueToName
		ctagModels = self.ctagModels
		msdModels = self.msdModels
		ctagFeatures = self.ctagFeatures
		msdFeatures = self.msdFeatures
		bestUnigramCtagForWord = self.bestUnigramCtagForWord
		bestUnigramMsdForWordCtagAttr = self.bestUnigramMsdForWordCtagAttr

		ctagCounter = 0
		onePossibleCtagCounter = 0
		correctCtagCounter = 0
		
		msdCounter = dict()
		onePossibleMsdCounter = dict()
		correctMsdCounter = dict()

		correctTagCounter = 0

		usingCtagUnigramModelCounter = 0
		usingMsdUnigramModelCounter = dict()
		noModelFound = 0

		ctagErrorsFound = list()
		ctagErrorsCounter = dict()
		
		msdErrorsFound = dict()
		msdErrorsCounter = dict()
		excessiveAttrsCounter = dict()									#counts predicted msd that where not in correct msd

		taggedText = list()
		parsedSegments = list()
		
		search = re.search
		while len(lines) > 0:
			#print "STARTING SENTENCE: ", line
			if line.find("<seg") >= 0:
				segment = dict()
				#print "STARTING SEGMENT: ", line					
				# inside segment
				#m = search('<seg xlink:href="(.+?)" xml:id="(.+?)">',line)
				#segmId = m.group(1)
				#morphId = m.group(2)
				# <f name = "orth">						
				while line.find('<f name="orth"') < 0:
					line = lines.pop()
				while line.find('<string>') < 0:
					line = lines.pop()
				m = search('<string>(.*?)</string>',line)
				segment['word'] = m.group(1)
				# <f name = "interps">
				while line.find('<f name="interps"') < 0:
					line = lines.pop()			
				# parse lex tags
				lexList = list()
				while True:
					while line.find('<fs type="lex" xml:id=') < 0 and line.find('<f name="disamb"') < 0:
						line = lines.pop()
					if line.find('<f name="disamb"') >= 0:
						break
					else:
						#    <fs type="lex" xml:id="morph_4.52.1-lex">
						m = search('<fs type="lex" xml:id="(.*?)"',line)
						xmlId = m.group(1)
						#       <f name="base">
						while line.find('<f name="base"') < 0:
							line = lines.pop()
						while line.find('<string>') < 0:
							line = lines.pop()
						m = search('<string>(.*?)</string>',line)
						base = m.group(1)
						#       <f name="ctag">
						while line.find('<f name="ctag"') < 0:
							line = lines.pop()
						while line.find('<symbol') < 0:
							line = lines.pop()
						m = search('value="(.*?)"',line)
						ctag = m.group(1)	
						#       <f name="msd">	
						while line.find('<f name="msd"') < 0:
							line = lines.pop()
						#			<symbol value=>
						while line.find('</f>') < 0:
							if line.find('<symbol') >= 0:
								m1 = search('value="(.*?)"',line) 
								m2 = search('xml:id="(.*?)"',line)
								lexList.append((base,ctag,m1.group(1),m2.group(1)))
							line = lines.pop()
						if len(lexList) < 1:
							print "ERROR: len(lexList) = 0 -> some <symbol> tags might have been not parsed"
				# <f name="disamb">
				while line.find('<f name="disamb"') < 0 and line.find('</seg>') < 0:
					line = lines.pop()
				if line.find('</seg>') >= 0:
					print 'ERROR: <f name="disamb"> not found'				
				while line.find("<f fVal=") < 0:
					line = lines.pop()
				m = search('<f fVal="#(.*?)" name="choice"/>', line)
				correctTagId = m.group(1)
				while line.find("</seg>") < 0:
					line = lines.pop()
				# we have reached end of segment
				for base,ctag,msd,msdId in lexList:
					if msdId == correctTagId:
						correctTag = (ctag,msd)
						correctCtag = ctag
				# create dictionary of correct msd attributes
				correctPossibleAttrs = set(pos[correctCtag] + pos_optional.get(correctCtag,[]))
				correctMsd = dict()
				for el in correctPossibleAttrs:					
					correctMsd[el] = u''
				splitMsd = correctTag[1].split(':')			
				for msdVal in splitMsd:
					if msdVal != u'':
						correctMsd[attrValueToName[msdVal]] = msdVal
				# find possible AC classes ctag and msd			
				possibleCtagValues = set()
				possibleMsdValues = dict()
				for (base,ctag,msd,xml_id) in lexList:
					possibleCtagValues.add(ctag)
					if ctag == correctCtag:
						foundAttrs = set()
						splitMsd = msd.split(':')
						for msdVal in splitMsd:
							if msdVal != u'':
								attrName = attrValueToName[msdVal]
								foundAttrs.add(attrName)
								attrSet = possibleMsdValues.setdefault(attrName,set())
								attrSet.add(msdVal)
						for attrName in correctPossibleAttrs.difference(foundAttrs):
							attrSet = possibleMsdValues.setdefault(attrName,set())
							attrSet.add(u'')
				possibleCtagsList = list(possibleCtagValues)
				possibleCtagsList.sort()
				possibleCtagsTuple = tuple(possibleCtagsList)
				possibleMsdValuesTuples = dict()
				for attrName, attrValuesSet in possibleMsdValues.iteritems():
					attrValuesList = list(attrValuesSet)
					attrValuesList.sort()
					attrValuesTuple = tuple(attrValuesList)
					possibleMsdValuesTuples[attrName] = attrValuesTuple	
				segment['msdDict'] = correctMsd
				segment['lexList'] = lexList
				segment['tag'] = correctTag
				segment['ctagAC'] = possibleCtagsTuple
				segment['msdAC'] = possibleMsdValuesTuples
				parsedSegments.append(segment)
			#if line.find("<seg") >= 0	
			else:
				line = lines.pop()
		#while

		prevPrev = dict()
		prev = dict()
		act = dict()
		next = dict()
		nextNext = dict()
		ctagAcTuplesNotFound = set()
		ctagAcTuplesFound = set()
		msdAcTuplesNotFound = dict()
		msdAcTuplesFound = dict()
		# process segments, find context, accuracy, errors, etc.
		parsedSegments.extend([dict(),dict()])
		for segment in parsedSegments:
			prevPrev = prev
			prev = act
			act = next
			next = nextNext
			nextNext = segment
			if len(act) > 0:
				#add context to ctag and msd ambiguity classes
				correctCtag = act['tag'][0]
				possibleCtagValuesTuple = act['ctagAC']
				# predict ctag value
				if len(possibleCtagValuesTuple) > 1:
					if possibleCtagValuesTuple in ctagModels:
						ctagAcTuplesFound.add(possibleCtagValuesTuple)
						ctagContext = ctagFeatures(prevPrev,prev,act,next,nextNext)	
						model = ctagModels[possibleCtagValuesTuple]
						predCtag = model.classify(ctagContext)
					else:
						ctagAcTuplesNotFound.add(possibleCtagValuesTuple)
						predCtag = bestUnigramCtagForWord.get(act['word'].lower(),u'')
						usingCtagUnigramModelCounter += 1
				else:
					predCtag = possibleCtagValuesTuple[0]
					onePossibleCtagCounter += 1 						
				ctagCounter += 1
				if correctCtag == predCtag:
					#sys.stdout.write("CORRECT CTAG %s : %s\n" % (correctCtag,predCtag))
					correctCtagCounter += 1
				else:
					#sys.stdout.write("WRONG CTAG %s : %s\n" % (correctCtag,predCtag))
					ctagErrorsFound.append((len(taggedText),correctCtag,predCtag))
				ctagErrorsCounterGivenCorrect = ctagErrorsCounter.setdefault(correctCtag,dict())
				oldCount = ctagErrorsCounterGivenCorrect.setdefault(predCtag,0)
				ctagErrorsCounterGivenCorrect[predCtag] = oldCount + 1
				#print 'correctCtag = ', correctCtag
				#print 'ctagErrorsCounterGivenCorrect = ',ctagErrorsCounterGivenCorrect
				#print (act['word'],correctCtag,predCtag,oldCount+1)
				# find possible msd values
				
				
				predMsd = dict()
				correctMsd = act['msdDict']
				msdAC = act['msdAC']
				possibleAttrs = set(pos.get(correctCtag,[]) + pos_optional.get(correctCtag,[]))
				if correctCtag == predCtag:
					# prepare msd unigram model
					word = act['word'].lower()
					bestUnigramMsdForAttr = dict()
					if word in bestUnigramMsdForWordCtagAttr:
						bestUnigramMsdForCtagAttr = bestUnigramMsdForWordCtagAttr[word]
						if predCtag in bestUnigramMsdForCtagAttr:
							bestUnigramMsdForAttr = bestUnigramMsdForCtagAttr[predCtag]
					# predict msd values
					msdModels2 = msdModels.get(predCtag,dict())						
					for attrName, attrValuesTuple in msdAC.iteritems():
						if len(attrValuesTuple) == 1:
							predMsd[attrName] = attrValuesTuple[0]
						elif attrName in msdModels2 and attrValuesTuple in msdModels2[attrName]:
							msdContext = msdFeatures(attrName,prevPrev,prev,act,next,nextNext)	
							model = msdModels2[attrName][attrValuesTuple]
							predAttr = model.classify(msdContext)
							predMsd[attrName] = predAttr
							#sys.stdout.write("MAXENT MODEL USED FOR %s : %s\n * FEATURES: %s\n" % (attrName,predAttr,msdContext))
							#sys.stdout.write(" * EXPLAIN: %s\n" % model.explain(msdContext))
							s = msdAcTuplesFound.setdefault(attrName,set()) 
							s.add(attrValuesTuple)
						else:
							s = msdAcTuplesNotFound.setdefault(attrName,set()) 
							s.add(attrValuesTuple)
							if attrName in bestUnigramMsdForAttr:
								predMsd[attrName] = bestUnigramMsdForAttr[attrName]	
								#sys.stdout.write("UNIGRAM MODEL USED FOR %s : %s\n" % (attrName,bestUnigramMsdForAttr[attrName]	))
								oldCount = usingMsdUnigramModelCounter.setdefault(attrName,0)
								usingMsdUnigramModelCounter[attrName] = oldCount + 1
							else:
								#sys.stdout.write("NO MODEL FOUND FOR %s\n" % attrName)
								noModelFound += 1
								predMsd[attrName] = u''
							#if
						#if
					#for
				#if correctCtag == predCtag
				
				for attrName in possibleAttrs:
					oldCount = msdCounter.setdefault(attrName,0)
					msdCounter[attrName] = oldCount + 1
					correctMsdAttrVal = correctMsd.get(attrName, u'')
					predMsdAttrVal = predMsd.get(attrName,u'')
					#sys.stdout.write(" | %s : correct = %s, pred = %s ... " % (attrName, correctMsdAttrVal,predMsdAttrVal))
					if predMsdAttrVal == correctMsdAttrVal:
						#sys.stdout.write("OK\n")
						oldCount = correctMsdCounter.setdefault(attrName,0)
						correctMsdCounter[attrName] = oldCount + 1
						if len(msdAC.get(attrName,())) == 1:
							oldCount = onePossibleMsdCounter.setdefault(attrName,0)
							onePossibleMsdCounter[attrName] = oldCount + 1
							#sys.stdout.write("ONE POSSIBLE MSD %s : %s\n" % (attrName,attrValuesTuple[0]))
					else:
						#sys.stdout.write("WRONG\n")		
						msdErrorsFoundGivenAttrName = msdErrorsFound.setdefault(attrName,list())
						msdErrorsFoundGivenAttrName.append( (len(taggedText), correctMsdAttrVal, predMsdAttrVal) )
					msdErrorsCounterGivenAttrName = msdErrorsCounter.setdefault(attrName,dict())
					msdErrorsCounterGivenAttrNameCorrect = msdErrorsCounterGivenAttrName.setdefault(correctMsdAttrVal,dict())
					oldCount = msdErrorsCounterGivenAttrNameCorrect.setdefault(predMsdAttrVal, 0)
					msdErrorsCounterGivenAttrNameCorrect[predMsdAttrVal] = oldCount + 1
				#excessiveAttrs = set(predMsd).difference(set(correctPossibleAttrs))
				#for attrName in excessiveAttrs:
				#	oldCount = excessiveAttrsCounter.setdefault(attrName,0)
				#	excessiveAttrsCounter[attrName] = oldCount + 1
				# Check if whole tag is correct
				if predCtag == correctCtag:
					tagOK = True
					for attrName in possibleAttrs:
						if predMsd.get(attrName,u'') != correctMsd.get(attrName, u''):
							tagOK = False
					if tagOK:
						correctTagCounter += 1
				predMsdString = ""
				for attrName in attrOrder:
					if attrName in predMsd and predMsd[attrName] != u'': 
						if predMsdString == "":
							predMsdString = predMsd[attrName]
						else:
							predMsdString = predMsdString+':'+predMsd[attrName]
				taggedText.append((act['word'],act['tag'],(predCtag,predMsdString)))
			#if len(act) > 0
		#for segment in parsedSegments.append([dict(),dict()]):							
			
		self.globalResults.append((
			filename,												# string
			len(taggedText),								# int
			onePossibleCtagCounter, 				# int
			correctCtagCounter, 						# int
			ctagCounter,										# int
			usingCtagUnigramModelCounter,		# int
			ctagErrorsCounter,							# correctCtag -> predCtag -> count (string -> string -> int)
			ctagErrorsFound,								# list of (num,correctCtag,predCtag)
			onePossibleMsdCounter,					# attrName -> count
			correctMsdCounter,							# attrName -> count
			msdCounter,											# attrName -> count
			usingMsdUnigramModelCounter,		# attrName -> count
			msdErrorsCounter,								# attrName -> correctVal -> predVal -> count
			msdErrorsFound,									# attrName -> list of (num,correctVal,predVal)
			excessiveAttrsCounter,					# attrName -> count
			noModelFound,										# int
			taggedText,											# list of tuples (word, correctTag, predTag)
			ctagAcTuplesFound,							# set of tuples
			ctagAcTuplesNotFound,						# set tuples
			msdAcTuplesFound,								# attrName -> set of tuples
			msdAcTuplesNotFound,						# attrName -> set of tuples
			correctTagCounter								# int, number of 100% correct tags
		))
		return taggedText
#		return (filename,segCount,correctCtagCounter,correctMsdCounter,gProb,errorsFound)


	################# CLEAR ###############
	def clear(self):

		self.ctagCounter = dict()
		self.msdCounter = dict()
		self.onePossibleCtag = 0
		self.onePossibleMsd = dict()
		self.unigramWordCtagCounter = dict()
		self.unigramWordCtagAttrMsdCounter = dict()
		self.unigramWordCtagProb = dict()
		self.unigramWordCtagAttrMsdProb = dict()
		self.bestUnigramCtagForWord = dict()
		self.bestUnigramMsdForWordCtagAttr = dict()
		self.msdTrainingCtagAttrClassesDict = dict()
		self.ctagTrainingClassesDict = dict()
		self.ctagModels = dict()
		self.msdModels = dict()
		self.globalResults = list()
		
	############# TEST ALL ###############

	def testAll(self,returnTaggedSents=False):
		print "########### TESTING MODEL ##########"
		t_start = time.time()
		taggedFiles = list()
		
		test = self.test
		testingFileList = self.testingFileList
		
		for filename in testingFileList:
			taggedFile = test(filename)
			if returnTaggedSents:
				taggedFiles.append((filename,taggedFile))
		
		t_end = time.time()
		print ""
		print "TESTING %d FILES TOOK %.2f SECONDS" % (len(testingFileList), t_end - t_start)

		if returnTaggedSents:
			return taggedFiles
		else:
			return []
			
	############# PRINT CHOSEN ERRORS #################
	
	def showErrors(self,attrName,correctVal,predVal):
		#codecs.ignore_errors(UnicodeEncodeError)
		globalResults = self.globalResults
		write = sys.stdout.write
		
		ctagErrorsFoundList = map(lambda x: (x[0],x[7]),globalResults)
		msdErrorsFoundList = map(lambda x: (x[0],x[13]),globalResults)
		taggedTextList = map(lambda x: (x[0],x[16]),globalResults)
		
		if attrName.lower() == "ctag":
			errorsList = ctagErrorsFoundList
		else:
			errorsList = map(lambda x: (x[0],x[1].get(attrName.lower(),list())), msdErrorsFoundList)
		
		
		selectedErrorsList = list()
		for index,(filename,errorPositions) in enumerate(errorsList):
			selectedErrors = list()
			for (position,correct,pred) in errorPositions:
				if correct == correctVal and pred == predVal:
					selectedErrors.append(position)
			selectedErrorsList.append((filename,index,selectedErrors))
		
		act = 0
		removedElements = list()
		
		print "############### INFO #################"
		print "word              correctTag | predTag"
		print "######################################"
		
		for (filename,index,selectedErrors) in selectedErrorsList:
			if len(selectedErrors) > 0:
				print "********** FILE " + filename + " **********"
			(taggedFilename,taggedText) = taggedTextList[index]
			maxPos = len(taggedText)
			if taggedFilename != filename:
				print "ERROR: filenames do not match: " + filename + ", " + taggedFilename
			for position in selectedErrors:
				start = max(0,position-7)
				end	= min(position+8,maxPos)
				word = taggedText[position][0]
				text = taggedText[start:end]
				write("----- %s -----\n" % word.encode('utf-8').upper())
				for (word,correctTag,predTag) in text:
					wordLen = len(word)
					fill = ' ' * (max(20-wordLen, 3))
					write("%s %s %s | %s\n" % (word.encode('utf-8'),fill,correctTag,predTag))
	
	############# PRINT SUMMARY ##############

	def printSummary(self):
	
		globalResults = self.globalResults
		pos = self.pos
		pos_optional = self.pos_optional
		attr = self.attr
		write = sys.stdout.write
		
		
		if len(globalResults) > 0:
			# filename,															# string
			# len(taggedText),											# int		
			# onePossibleCtagCounter, 							# int
			# correctCtagCounter, 									# int
			# ctagCounter,													# int
			# usingCtagUnigramModelCounter,					# int
			# ctagErrorsCounter,										# correctCtag -> predCtag -> count 
			# ctagErrorsFound,											# list of (num,word,correctCtag,predCtag)
			# onePossibleMsdCounter,								# attrName -> count
			# correctMsdCounter,										# attrName -> count
			# msdCounter,														# attrName -> count
			# usingMsdUnigramModelCounter,					# attrName -> count
			# msdErrorsCounter,											# attrName -> correctVal -> predVal -> count
			# msdErrorsFound,												# attrName -> list of (num,word,correctMsd,predMsd)
			# excessiveAttrsCounter,								# attrName -> count
			# noModelFound													# int
			# taggedText														# list of tuples (word, correctTag, predTag)
			# ctagAcTuplesFound,										# set of tuples
			# ctagAcTuplesNotFound,									# set tuples
			# msdAcTuplesFound,											# attrName -> set of tuples
			# msdAcTuplesNotFound										# attrName -> set of tuples
			#	correctTagCounter											# int, number of 100% correct tags
			
			onePossibleCtagCounter = sum(map(lambda x: x[2],globalResults))
			correctTagCounter = sum(map(lambda x : x[21], globalResults))
			correctCtagCounter = sum(map(lambda x: x[3],globalResults))
			ctagCounter = sum(map(lambda x: x[4],globalResults))
			usingCtagUnigramModelCounter = sum(map(lambda x: x[5],globalResults))
			
			ctagErrorsCounterList = map(lambda x: x[6],globalResults)


			onePossibleMsdCounterList = map(lambda x: x[8],globalResults)
			correctMsdCounterList = map(lambda x: x[9],globalResults)
			msdCounterList = map(lambda x: x[10],globalResults)
			usingMsdUnigramModelCounterList = map(lambda x: x[11],globalResults)
			msdErrorsCounterList = map(lambda x: x[12],globalResults)
			excessiveAttrsCounterList = map(lambda x: x[14],globalResults)
			
			
			noModelFound = sum(map(lambda x: x[15],globalResults))

			ctagAcTuplesFound =	set()								# set of tuples
			ctagAcTuplesNotFound = set()						# set tuples
			msdAcTuplesFound = dict()							# attrName -> set of tuples
			msdAcTuplesNotFound = dict()						# attrName -> set of tuples


			# find AC tuples found and not found in the model
			for res in globalResults:
				ctagAcTuplesFound.update(res[17])
				ctagAcTuplesNotFound.update(res[18])
				for attrName,foundSet in res[19].iteritems():
					s = msdAcTuplesFound.setdefault(attrName,set())
					s.update(foundSet)
				for attrName,notFoundSet in res[20].iteritems():
					s = msdAcTuplesNotFound.setdefault(attrName,set())
					s.update(notFoundSet)
			
			#print "%%%%%%%%%%%%%%%%%%%%%%%%%%% CTAG AC TUPLES FOUND: "
			#pprint.pprint(ctagAcTuplesFound)

			#print "%%%%%%%%%%%%%%%%%%%%%%%%%%% CTAG AC TUPLES NOT FOUND: "
			#pprint.pprint(ctagAcTuplesNotFound)

			#print "%%%%%%%%%%%%%%%%%%%%%%%%%%% MSD AC TUPLES FOUND: "
			#pprint.pprint(msdAcTuplesFound)

			#print "%%%%%%%%%%%%%%%%%%%%%%%%%%% MSD AC TUPLES NOT FOUND: "
			#pprint.pprint(msdAcTuplesNotFound)
		
			#print "%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%% "
		
					
			ctagErrorsCounter = dict()			
			onePossibleMsdCounter = dict()
			correctMsdCounter = dict()
			msdCounter = dict()
			usingMsdUnigramModelCounter = dict()
			msdErrorsCounter = dict()
			excessiveAttrsCounter = dict()
		
			for res in ctagErrorsCounterList:
				for (correct, predCountDict) in res.iteritems():
					ctagErrorsCounterGivenCorrect = ctagErrorsCounter.setdefault(correct,dict())
					for (pred,count) in predCountDict.iteritems():
						oldCount = ctagErrorsCounterGivenCorrect.setdefault(pred,0)
						ctagErrorsCounterGivenCorrect[pred] = oldCount + count						

			for res in msdErrorsCounterList:
				for (attrName,correctPredCountDict) in res.iteritems():
					msdErrorsCounterGivenAttrName = msdErrorsCounter.setdefault(attrName,dict())
					for (correct, predCountDict) in correctPredCountDict.iteritems():
						msdErrorsCounterGivenAttrNameCorrect = msdErrorsCounterGivenAttrName.setdefault(correct,dict())
						for (pred,count) in predCountDict.iteritems():
							msdErrorsCounterGivenAttrNameCorrect.setdefault(pred,0) 
							msdErrorsCounterGivenAttrNameCorrect[pred] += count 

			for res in excessiveAttrsCounterList:
				for (attrName,count) in res.iteritems():
					oldCount = excessiveAttrsCounter.setdefault(attrName,0)
					excessiveAttrsCounter[attrName] = oldCount + count


			for res in onePossibleMsdCounterList:
				for (attrName,count) in res.iteritems():
					oldCount = onePossibleMsdCounter.setdefault(attrName,0)
					onePossibleMsdCounter[attrName] = oldCount + count
			for res in correctMsdCounterList:
				for (attrName,count) in res.iteritems():
					oldCount = correctMsdCounter.setdefault(attrName,0)
					correctMsdCounter[attrName] = oldCount + count
			for res in msdCounterList:
				for (attrName,count) in res.iteritems():
					oldCount = msdCounter.setdefault(attrName,0)
					msdCounter[attrName] = oldCount + count
			for res in usingMsdUnigramModelCounterList:
				for (attrName,count) in res.iteritems():
					oldCount = usingMsdUnigramModelCounter.setdefault(attrName,0)
					usingMsdUnigramModelCounter[attrName] = oldCount + count
			
			trainingCtagCount = sum(self.ctagCounter.itervalues())
			trainingMsdCount = self.msdCounter
			print "#############################"
			print "TRAINING STATISTICS "
			print "  LEARNING FILES \t%d" % len(self.learningFileList)	
			print "  CTAG ONE POSSIBLE VALUE \t%d\t%d\t%.2f %%" % (self.onePossibleCtag, trainingCtagCount, 100.0 * float(self.onePossibleCtag) / float(trainingCtagCount) )			
			print "  MSD ONE POSSIBLE VALUE " 
			for (attrName, val) in self.onePossibleMsd.iteritems():
				attrCount = trainingMsdCount[attrName]
				if attrCount > 0:
					percent = 100.0 * float(val) / float(attrCount)
				else:
					percent = 0.0
				print "    %s\t%d\t%d\t%.2f %%" % (attrName, val, attrCount, percent)
			print "#############################"
			print "TESTING RESULTS" 
			print "  TESTING FILES \t%d" % len(self.testingFileList)
			print "  TAG ACCURACY \t%d\t%d\t%.2f %%" % (correctTagCounter, ctagCounter, 100.0 * float(correctTagCounter) / float(ctagCounter))
			print "  CTAG ACCURACY \t%d\t%d\t%.2f %%" % (correctCtagCounter, ctagCounter, 100.0 * float(correctCtagCounter) / float(ctagCounter))
			print "  CTAG ONE POSSIBLE VALUE \t%d\t%d\t%.2f %%" % (onePossibleCtagCounter, ctagCounter, 100.0 * float(onePossibleCtagCounter) / float(ctagCounter))
			print "  CTAG UNIGRAM MODEL USED \t%d\t%d\t%.2f %%" % (usingCtagUnigramModelCounter, ctagCounter, 100.0 * float(usingCtagUnigramModelCounter) / float(ctagCounter))
			print "  MSD ACCURACY: "
			for (attrName,val) in correctMsdCounter.iteritems():
				attrCount = msdCounter[attrName]
				if attrCount > 0:
					percent = 100.0 * float(val) / float(attrCount)
				else:
					percent = 0.0			
				print "    %s\t%d\t%d\t%.2f %%" % (attrName, val, attrCount, percent)	
			print "  MSD ONE POSSIBLE VALUE "
			for (attrName,val) in onePossibleMsdCounter.iteritems():
				attrCount = msdCounter[attrName]
				if attrCount > 0:
					percent = 100.0 * float(val) / float(attrCount)
				else:
					percent = 0.0			
				print "    %s\t%d\t%d\t%.2f %%" % (attrName, val, attrCount, percent)	
			print "  MSD NO MODEL FOUND \t%d" % noModelFound
			print "  MSD UNIGRAM MODEL USED	"
			for (attrName,val) in usingMsdUnigramModelCounter.iteritems():
				attrCount = msdCounter[attrName]
				if attrCount > 0:
					percent = 100.0 * float(val) / float(attrCount)
				else:
					percent = 0.0							
				print "    %s\t%d\t%d\t%.2f %%" % (attrName, val, attrCount, percent)	

			print "#############################"
			print "EXCESSIVE ATTRS"
			for (attrName,count) in excessiveAttrsCounter.iteritems():
				print "  %s\t%d" % (attrName,count)
		
			print "#############################"
			print "CTAG CONTINGENCY TABLE"
			write("\tCOUNT")
			for name in pos.keys() + [u'']:
				write("\t%s" % name)
			write("\n")			
			for correct in pos:
				ctagErrorsCounterGivenCorrect = ctagErrorsCounter.get(correct,dict())
				ctagCount = sum(ctagErrorsCounterGivenCorrect.itervalues())
				write("%s\t%d\t" % (correct,ctagCount))
				for pred in pos.keys() + [u'']:
					valCount = float(ctagErrorsCounterGivenCorrect.get(pred,0))
					if ctagCount > 0 and valCount > 0.0:
						percent = (100.0 * valCount / float(ctagCount))
						write("%.2f %%\t" % percent)
					else:
						write("\t")
				write("\n")

			print "#############################"
			print "MSD CONTINGENCY TABLES "
			
			optionalAttrs = set()
			for attrList in pos_optional.itervalues():
				optionalAttrs.update(attrList)
			for attrName in attr:
				attrValues = attr[attrName]
				attrValues.append(u'')
				msdErrorsCounterGivenAttrName = msdErrorsCounter.get(attrName,dict())
				print "CONTINGENCY TABLE FOR %s" % attrName	
				write("\tCOUNT")
				for attrVal in attrValues:
					write("\t%s" % attrVal)
				write("\n")
				for correct in attrValues:
					msdErrorsCounterGivenAttrNameCorrect = msdErrorsCounterGivenAttrName.get(correct,dict())
					attrCount = 0
					for pred in attrValues:
						attrCount += msdErrorsCounterGivenAttrNameCorrect.get(pred,0)
					attrCount = float(attrCount)
					write("%s\t%d\t" % (correct,attrCount)) 
					for pred in attrValues:
						valCount = float(msdErrorsCounterGivenAttrNameCorrect.get(pred,0))
						if attrCount > 0.0 and valCount > 0.0:					
							percent = 100.0 * valCount / attrCount
							write("%.2f %%\t" % percent)
						else:							
							write("\t")
					write("\n")
		else:
			print "   SUMMARY EMPTY"

				
	############ SAVE MODEL #################

	def saveModel(self):
		t_start = time.time()
		# SAVE ALL DATA
		mDir = os.path.join(self.testingDir,"_model")

		shutil.rmtree(mDir, ignore_errors=True)
		os.mkdir(mDir)

		for varname in self.varList:
			if hasattr(self,varname):
				f = open(os.path.join(mDir,varname),"w")
				cPickle.dump(getattr(self,varname),f)
				f.close()

		if hasattr(self,'testingDir') and hasattr(self,'learningDirList'):
			file = open(os.path.join(mDir,"readme.txt"),'w')
			file.write("FILENAME: "+mDir+"\n")
			file.write("TIME: "+time.ctime()+"\n")
			file.write("testingDir: " + self.testingDir + "\n")
			file.write("learningDirList: ")
			for f in self.learningDirList:
				file.write("  "+f)
			file.close()
		t_end = time.time()
		print "SAVING MODEL TOOK %.2f SECONDS" % (t_end - t_start)

	############ LOAD MODEL ############

	def loadModel(self,varNameList=[]):
		t_start = time.time()

		# LOAD DATA
		mDir = os.path.join(self.testingDir,"_model")

		if varNameList == []:	
			varList = self.varList
		else:
			varList = varNameList
				
		for varname in varList:
			if os.path.isfile(os.path.join(mDir,varname)):		
				f = open(os.path.join(mDir,varname),"r")
				setattr(self,varname,cPickle.load(f))
				f.close()
	
		if os.path.isfile(os.path.join(mDir,"readme.txt")):					
			file = open(os.path.join(mDir,"readme.txt"),'r')
			info = file.read()
			print info
			file.close()
			
		t_end = time.time()
		print "LOADING MODEL TOOK %.2f SECONDS" % (t_end - t_start)
	
	
