#!/bin/bash
#autor: Radomir Mastalerz

# Run 10-fold cross-validation. Results will be printed on standard output.

startTime=`date +%s`
for subset in `seq 0 9`; do
	echo "====================================== CROSS VALIDATION $subset ===================================="
	root=`pwd`
	echo "ROOT = " $root
	##### TWORZENIE KATALOGOW DOCELOWYCH TYMCZASOWYCH #####
	if [ -d "$root/out/$subset" ]; then
		rm -r "$root/out/$subset"
	fi
	mkdir $root"/out/"$subset
		n=0
	for mach in `cat machines`; do
		machine[$n]=$mach
		out_file[$n]=$root"/out/"$subset"/"$mach"_"$n
		n=$[$n+1]
	done
	
	##### PREPARATION #############
	python $root/preDistributed.py $subset &
	pid=$!
	wait $pid
	ret=$?
	if [ $ret -eq 0 ]; then
		sleep 5
	else
		echo "PRE DISTRIBUTED ERROR, EXITING WITH STATUS $ret"
		exit
	fi
	
	##### DISTRIBUTED LEARNING #####
	for k in `seq 0 $[$n-1]`; do
		ssh ${machine[$k]} python $root/learnSelectedModels.py $root $subset $k $n 1>>${out_file[$k]} 2>>${out_file[$k]} & 
		ssh_pid[$k]=$!
	done    
	
	sleep 5
	while [ ${#ssh_pid[*]} -gt 0 ]; do
		echo "## NEW WAIT LOOP"
		for k in ${!ssh_pid[*]}; do
			echo -n "$k WAITING FOR ${machine[$k]} "
			wait ${ssh_pid[$k]}
			ret_value=$?
			if [ $ret_value -eq 0 ]; then
				echo "OK"
				unset ssh_pid[$k]
			else
				if [ $ret_value -eq 255 ]; then								#error connecting through ssh
					echo "FAILED $ret_value"
				else
					echo "ERR $ret_value"
				fi
				newMachineIndex=$[$RANDOM%$n]
				newMachine=${machine[$newMachineIndex]}
				echo "RERUNNING $k ON MACHINE $newMachine"
				ssh $newMachine python $root/learnSelectedModels.py $root $subset $k $n 1>>${out_file[$k]}"_"$newMachine 2>>${out_file[$k]}"_"$newMachine &
				ssh_pid[$k]=$!
				machine[$k]=$newMachine
			fi
		done		
	done
       
	#### COMBINING RESULTS ######
	if [ "$cont" == "NO" ]; then
		echo "ERROR FOUND, NOT RUNNING postDistributed.py"
		echo "EXITING"
		exit
	else
		sleep 5
		python $root/postDistributed.py $subset $n &
		pid=$!
		wait $pid
		ret=$?
		echo "RETURN VALUE = $ret"
		if [ $ret -eq 0 ]; then
			echo "RETURN VALUE OK"
			sleep 5
		else
			echo "POST DISTRIBUTED ERROR, EXITING"
			exit
		fi
	fi
done

endTime=`date +%s`
echo "TOTAL CROSS VALIDATION RUNNING TIME = $[$endTime-$startTime]"


